﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Linq.Expressions;
using System.Collections;
using System.Reflection;

using IdeaBlade.EntityModel;
using IdeaBlade.Linq;
using DomainModel;

namespace WpfConsole {

  public class MainDemo {
    ResultsReporter _resultsReporter;

    public MainDemo(ResultsReporter callbackMethod) {
      Utility.InitializeDemo(_nameDemo, ref callbackMethod, ref _resultsReporter, ref _localOutput);
    }

    #region Portals

    public void Concurrency_RemoteModifyLocalModify() {
      WriteToOutput("[{0}] Starting...\n", _nameConcurrency_RemoteModifyLocalModify);
      ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteModifyLocalModify);
    }

    public void Concurrency_RemoteModifyLocalDelete() {
      WriteToOutput("[{0}] Starting...\n", _nameConcurrency_RemoteModifyLocalDelete);
      ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteModifyLocalDelete);
    }

    public void Concurrency_RemoteDeleteLocalModify() {
      WriteToOutput("[{0}] Starting...\n", _nameConcurrency_RemoteDeleteLocalModify);
      ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteDeleteLocalModify);
    }

    public void Concurrency_RemoteDeleteLocalDelete() {
     WriteToOutput("[{0}] Starting...\n", _nameConcurrency_RemoteDeleteLocalDelete);
      ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteDeleteLocalDelete);
    }

    #endregion Portals

    #region ConcurrencyConflictSimulator

    private void ConcurrencyConflictSimulator(ConcurrencyScenario scenario) {
      string nameOfExecutingTest = GetOutputLabelForExecutingTest(scenario);
      WriteToOutput("\n\n[{0}] Started...\n", nameOfExecutingTest);

      NorthwindIBModelManager emLocal = new NorthwindIBModelManager(); // represents environment of local user
      NorthwindIBModelManager emRemote = new NorthwindIBModelManager(); // represents environment of remote user

      // Create and save a new Customer and a related Order. 
      // We're doing this instead of retrieving existing data simply in order
      // to leave the latter untouched by this demo.
      Customer theCustomerInstanceLocal = CreateAndSaveCustomerAndRelatedOrder(emLocal);

      // Get a second instance of the same Customer into a different EntityManager.
      // This simulates the scenario of a second User somewhere across the building 
      // or across the planet working with the same Customer record.
      Customer theCustomerInstanceRemote = theCustomerInstanceLocal.ToQuery().With(emRemote).FirstOrNullEntity();

      ModifyInstances(scenario, theCustomerInstanceLocal, theCustomerInstanceRemote);

      SavePendingChanges(nameOfExecutingTest, emRemote);
      SavePendingChanges(nameOfExecutingTest, emLocal);

      Utility.FlushOutputBuffer(_localOutput, _resultsReporter);
    }

    private static Customer CreateAndSaveCustomerAndRelatedOrder(NorthwindIBModelManager em) {
      Customer theCustomerInstance1 = new Customer();
      theCustomerInstance1.CustomerID = System.Guid.NewGuid();
      theCustomerInstance1.CompanyName = "theCustomerInstance1";
      em.AddEntity(theCustomerInstance1);

      Order newOrder1 = new Order();
      em.AddEntity(newOrder1);
      theCustomerInstance1.Orders.Add(newOrder1);

      em.SaveChanges();

      return theCustomerInstance1;
    }

    /// <summary>
    /// Save the changed entities if possible, even if this requires
    /// overriding a concurrency conflict.
    /// </summary>
    /// <param name="nameOfExecutingTest"></param>
    /// <param name="em"></param>
    /// <remarks>
    /// Demonstrates techniques for overriding a concurrency conflict under
    /// different conflict scenarios. In a production app, you would not do
    /// that as a matter of course; on the other hand, you might wish to do it 
    /// under certain business scenarios. This shows you how to do it if you need to.
    /// </remarks>
    private void SavePendingChanges(string nameOfExecutingTest, NorthwindIBModelManager em) {
      WriteToOutput(nameOfExecutingTest, em);

      bool saveSucceeded = false;
      while (!saveSucceeded) {
        if (_unsuccessfulSaveAttempts > _MAX_UNSUCCESSFUL_SAVE_ATTEMPTS) {
          string msg = "[{0}] Exceeded limit ({1}) on unsuccessful save attempts!\n"
            + "This could occur if the number of conflicted entities submitted\n"
            + "for saving simply exceeds that limit, or if the attempt to ready\n"
            + "a conflicted entity for saving isn't working!\n";
          WriteToOutput(msg, nameOfExecutingTest, _MAX_UNSUCCESSFUL_SAVE_ATTEMPTS);
          break;
        }

        try {
          SaveResult theSaveResult = em.SaveChanges();
          if (theSaveResult.Ok) {
            saveSucceeded = true;
            Entity entitySaved = (Entity)theSaveResult.SavedEntities[0];
          } else {
            _unsuccessfulSaveAttempts++;

            if (theSaveResult.WasExceptionHandled) {
              WriteToOutput("[{0}] Problem saving: {1}\n", nameOfExecutingTest, 
                theSaveResult.HandledException.Message);
            } else {
              WriteToOutput("[{0}] Problem saving: Exception was not handled\n", nameOfExecutingTest);
            }
          }
        }
        catch (Exception ex) {
          if (ex is EntityManagerSaveException) {
            // Identify the concurrency scenario...
            Entity entityWithError = (Entity)(ex as EntityManagerSaveException).EntitiesWithErrors[0];

            string msg = "[{0}] Error on attempt to save: {1}\n\t{2}\n\n" +
              "\tApplying fix to permit forced save...\n";
            WriteToOutput(msg, nameOfExecutingTest, entityWithError.EntityAspect.EntityKey.ToString(),
              ex.Message);

            // Get a copy of the target entity from the database, if it still exists there...
            List<Object> remoteEntities = GetEntitiesFromDb(entityWithError);

            // The code in the following switch construct takes action to put the conflicted
            // entity in a state such that it will be saved successfully if submitted again (as
            // will happen in this save loop). As such, it causes the app to behave as if
            // concurrency checking were not even activated (Last One In Wins).
            //
            // In a production app, having bothered to check for concurrency conflicts in the first
            // place, you would only conditionally execute this logic to override the conflict. 
            // For example, you might do so after notifying the end user of the conflict and 
            // receiving an instruction from her to perform the save in spite of it; or you might 
            // have business logic that decides when to force the save (or when to permit the end 
            // user to decide).

            EntityState state = entityWithError.EntityAspect.EntityState;
            switch (state) {
              case EntityState.Deleted:
                // The "local" instance was marked for deletion.
                // Since we got a concurrency conflict, the remote entity was either
                // modified or deleted. If it exists, it was modified.

                if (remoteEntities.Count == 0) {
                  // Discovered scenario: ConcurrencyScenario.RemoteDeleteLocalDelete
                  saveSucceeded = true;
                  // No further action is required. The local user wanted the record deleted,
                  // and it has been deleted, albeit by the remote user. Since we know all of
                  // this, we could inform the local user about exactly what happened, if
                  // we deem that important to do.
                } else {
                  // Discovered scenario: ConcurrencyScenario.RemoteModifyLocalDelete
                  em.RefetchEntity(entityWithError, MergeStrategy.PreserveChangesUpdateOriginal);
                }
                break;

              case EntityState.Modified:
                // The "local" instance was modified.
                // Since we got a concurrency conflict, the remote entity was either
                // modified or deleted. If it exists, it was modified.

                if (remoteEntities.Count == 0) {
                  // Discovered scenario: ConcurrencyScenario.RemoteDeleteLocalModify
                  entityWithError.EntityAspect.SetAdded();
                } else {
                  // Discovered scenario: ConcurrencyScenario.RemoteModifyLocalModify
                  em.RefetchEntity(entityWithError, MergeStrategy.PreserveChangesUpdateOriginal);
                }
                break;

              default:
                throw new Exception("Unexpected EntityState: " + state.ToString());
            }
          } else {
            WriteToOutput("[{0}] Error on save attempt: {1}\n", nameOfExecutingTest, ex.Message);            
          }
        }
      }
      if (saveSucceeded) {
        WriteToOutput("[{0}] Saved succeeded!\n", nameOfExecutingTest);
      }
    }

    private static List<Object> GetEntitiesFromDb(Entity entity) {
      NorthwindIBModelManager emTemp = new NorthwindIBModelManager();
      EntityKeyQuery entityQuery = entity.EntityAspect.EntityKey.ToKeyQuery();
      List<Object> entities = entityQuery.With(emTemp).Execute().ToList();
      return entities;
    }

    private static void ModifyInstances(ConcurrencyScenario scenario,
      Customer theCustomerInstanceLocal, Customer theCustomerInstanceRemote) {
      switch (scenario) {
        case ConcurrencyScenario.RemoteModifyLocalModify:
          ModifyCustomer(theCustomerInstanceRemote, "R");
          ModifyCustomer(theCustomerInstanceLocal, "L");
          break;
        case ConcurrencyScenario.RemoteModifyLocalDelete:
          ModifyCustomer(theCustomerInstanceRemote, "R");
          MarkCustomerForDeletion(theCustomerInstanceLocal);
          break;
        case ConcurrencyScenario.RemoteDeleteLocalModify:
          MarkCustomerForDeletion(theCustomerInstanceRemote);
          ModifyCustomer(theCustomerInstanceLocal, "L");
          break;
        case ConcurrencyScenario.RemoteDeleteLocalDelete:
          MarkCustomerForDeletion(theCustomerInstanceRemote);
          MarkCustomerForDeletion(theCustomerInstanceLocal);
          break;
        default:
          throw new Exception("Unknown ConcurrencyScenario. Aborting...");
      }
    }

    private static void ModifyCustomer(Customer theCustomerInstance, string suffix) {
      theCustomerInstance.CompanyName += suffix;
    }

    private static void MarkCustomerForDeletion(Customer theCustomerInstance) {

      // Related Orders must be deleted before we can delete this Customer...
      List<Order> orders = theCustomerInstance.Orders.ToList();
      for (int i = orders.Count - 1; i > -1; i--) {
        orders[i].EntityAspect.Delete();
      }

      theCustomerInstance.EntityAspect.Delete();
    }

    private void WriteToOutput(string nameOfExecutingTest, NorthwindIBModelManager em) {
      WriteToOutput("[{0}] Entities to be saved...\n", nameOfExecutingTest);

      foreach (Entity anEntity in em.FindEntities(EntityState.AnyAddedModifiedOrDeleted)) {
        WriteToOutput("\t{0}\n", anEntity.EntityAspect.EntityKey.ToString());
      }
    }

    private void WriteToOutput(string message, params object[] args) {
      string msg = "";
      if (args.Length == 1) {
        msg = string.Format(message, args[0]);
      } else if (args.Length == 2) {
        msg = string.Format(message, args[0], args[1]);
      } else if (args.Length == 3) {
        msg = string.Format(message, args[0], args[1], args[2]);
      }
      _localOutput.Append(msg);
    }

    private static string GetOutputLabelForExecutingTest(ConcurrencyScenario scenario) {
      string nameOfExecutingTest;
      switch (scenario) {
        case ConcurrencyScenario.RemoteModifyLocalModify:
          nameOfExecutingTest = "RemoteModify/LocalModify";
          break;
        case ConcurrencyScenario.RemoteModifyLocalDelete:
          nameOfExecutingTest = "RemoteModify/LocalDelete";
          break;
        case ConcurrencyScenario.RemoteDeleteLocalModify:
          nameOfExecutingTest = "RemoteDelete/LocalModify";
          break;
        case ConcurrencyScenario.RemoteDeleteLocalDelete:
          nameOfExecutingTest = "RemoteDelete/LocalDelete";
          break;
        default:
          nameOfExecutingTest = "";
          throw new Exception("Unknown ConcurrencyScenario. Aborting...");
      }
      return nameOfExecutingTest;
    }

    #endregion ConcurrencyConflictSimulator

    #region Private Fields

    const string _nameDemo = "MainDemo (MDM)";
    const string _nameConcurrency_RemoteModifyLocalModify = "MDM:Concurrency_RemoteModifyLocalModify";
    const string _nameConcurrency_RemoteDeleteLocalModify = "MDM:Concurrency_RemoteDeleteLocalModify";
    const string _nameConcurrency_RemoteModifyLocalDelete = "MDM:Concurrency_RemoteModifyLocalDelete";
    const string _nameConcurrency_RemoteDeleteLocalDelete = "MDM:Concurrency_RemoteDeleteLocalDelete";

    StringBuilder _localOutput;

    private enum ConcurrencyHandlerStrategy {
      SetAdded,
      Refetch
    }

    private enum ConcurrencyScenario {
      RemoteModifyLocalModify,
      RemoteModifyLocalDelete,
      RemoteDeleteLocalModify,
      RemoteDeleteLocalDelete
    }

    int _unsuccessfulSaveAttempts = 0;
    const int _MAX_UNSUCCESSFUL_SAVE_ATTEMPTS = 10;

    #endregion Private Fields

  }
}
